@@ -2,7 +2,7 @@
 from tinydoc import *
 import os
 
-#os.system("dot -Tpng docs/depgraph.dot > docs/depgraph.png")
+os.system("dot -Tpng docs/depgraph.dot > docs/depgraph.png")
 
 
 info = DocInfo()
@@ -22,6 +22,7 @@ root.nodes += [ParseTextFile(Node("Custom Actions"), "docs/actions.txt")]
 root.nodes += [ParseFile(Node("Command Line Reference"), "src/main.c")]
 root.nodes += [ParseFile(Node("Function Reference"), "src/base.lua").Sorted()]
 root.nodes += [ParseFile(Node("Tool Reference"), "src/tools.lua").Sorted()]
+root.nodes += [ParseTextFile(Node("Quirks"), "docs/quirks.txt")]
 root.nodes += [ParseTextFile(Node("License"), "license.txt", True)]
 #notes.nodes +=[Node("C/C++ Dependency Checker")]
 #notes.nodes +=[Node("Spaces in Paths")]
