@@ -241,11 +241,15 @@ test("collect_recurse")
 test("sharedlib")
 test("deadlock")
 test("addorder")
+test("import")
+test("multipleoutput")
 
 if len(failed_tests):
 	print "FAILED TESTS:"
 	for t in failed_tests:
 		print "\t"+t
+	sys.exit(1)
 else:
 	print "ALL TESTS PASSED!"
+	sys.exit(0)
 
