@@ -105,25 +105,33 @@ struct CACHERUNINFO
 static int dependency_cpp_run(struct CONTEXT *context, struct NODE *node,
 		int (*callback)(struct NODE *, void *, const char *, int), void *userdata);
 
-static void cachehit_callback(struct NODE *node, struct CACHENODE *cachenode, void *user)
+static void cachehit_callback(struct NODE *node, struct CACHEINFO_DEPS *cacheinfo, void *user)
 {
 	struct CACHERUNINFO *info = (struct CACHERUNINFO *)user;
 	
 	/* check if the file has been removed */
-	struct NODE *existing_node = node_find_byhash(node->graph, cachenode->hashid);
+	struct NODE *existing_node = node_find_byhash(node->graph, cacheinfo->hashid);
 	if(existing_node)
 	{
-		struct NODE *newnode = node_add_dependency_withnode(node, existing_node);
+		struct NODE *newnode = node_add_dependency (node, existing_node);
 		dependency_cpp_run(info->context, newnode, info->callback, info->userdata);
 	}
 	else
 	{
-		if(file_timestamp(cachenode->filename) == 0)
-			node->dirty = NODEDIRTY_MISSING;
+		time_t timestamp = file_timestamp(cacheinfo->filename);
+		if(timestamp)
+		{
+			/* this shouldn't be able to fail */
+			struct NODE *newnode;
+			node_create(&newnode, info->context->graph, cacheinfo->filename, NULL, timestamp);
+			node_add_dependency (node, newnode);
+
+			/* recurse the dependency checking */
+			dependency_cpp_run(info->context, newnode, info->callback, info->userdata);
+		}
 		else
 		{
-			struct NODE *newnode = node_add_dependency(node, cachenode->filename);
-			dependency_cpp_run(info->context, newnode, info->callback, info->userdata);
+			node->dirty = NODEDIRTY_MISSING;
 		}
 	}
 }
@@ -159,7 +167,7 @@ static int dependency_cpp_run(struct CONTEXT *context, struct NODE *node,
 	cacheinfo.context = context;
 	cacheinfo.callback = callback;
 	cacheinfo.userdata = userdata;
-	if(cache_do_dependency(context, node, cachehit_callback, &cacheinfo))
+	if(depcache_do_dependency(context, node, cachehit_callback, &cacheinfo))
 		return 0;
 
 	/* mark the node as checked */
@@ -178,7 +186,7 @@ static int dependency_cpp_run(struct CONTEXT *context, struct NODE *node,
 	
 	if(!filebuf)
 	{
-		printf("cpp-dep: error allocating %ld bytes\n", filesize);
+		printf("cpp-dep: %s: error allocating %ld bytes\n", node->filename, filesize);
 		fclose(file);
 		return 1;
 	}
@@ -189,7 +197,7 @@ static int dependency_cpp_run(struct CONTEXT *context, struct NODE *node,
 
 	if(readitems != filesize)
 	{
-		printf("cpp-dep: error reading the complete file. %ld of %ld bytes read\n", readitems, filesize);
+		printf("cpp-dep: %s: error reading. %ld of %ld bytes read\n", node->filename, readitems, filesize);
 		free(filebuf);
 		return 1;
 	}
@@ -228,15 +236,33 @@ struct CPPDEPINFO
 	struct STRINGLIST *paths;
 };
 
+static int node_findfile(struct GRAPH *graph, const char *filename, struct NODE **node, time_t *timestamp)
+{
+	/* first check the graph */
+	*node = node_find(graph, filename);
+	if(*node)
+		return 1;
+
+	/* then check the file system */
+	*timestamp = file_timestamp(filename);
+	if(*timestamp && file_isregular(filename))
+		return 1;
+
+	return 0;
+}
+
 /* */
 static int dependency_cpp_callback(struct NODE *node, void *user, const char *filename, int sys)
 {
 	struct CPPDEPINFO *depinfo = (struct CPPDEPINFO *)user;
-	struct NODE *depnode;
 	struct CPPDEPINFO recurseinfo;
 	char buf[MAX_PATH_LENGTH];
 	int check_system = sys;
+
 	int found = 0;
+	struct NODE *depnode = NULL;
+	time_t timestamp = 0;
+	
 	
 	if(!sys)
 	{
@@ -250,7 +276,7 @@ static int dependency_cpp_callback(struct NODE *node, void *user, const char *fi
 		}
 		path_join(node->filename, flen, filename, -1, buf, sizeof(buf));
 		
-		if(file_exist(buf) || node_find(node->graph, buf))
+		if(node_findfile(node->graph, buf, &depnode, &timestamp))
 			found = 1;
 		else
 		{
@@ -264,7 +290,7 @@ static int dependency_cpp_callback(struct NODE *node, void *user, const char *fi
 		/* <system.header> */
 		if(path_isabs(filename))
 		{
-			if(file_exist(filename) || node_find(node->graph, filename))
+			if(node_findfile(node->graph, filename, &depnode, &timestamp))
 			{
 				strcpy(buf, filename);
 				found = 1;
@@ -278,7 +304,7 @@ static int dependency_cpp_callback(struct NODE *node, void *user, const char *fi
 			for(cur = depinfo->paths; cur; cur = cur->next)
 			{
 				path_join(cur->str, cur->len, filename, flen, buf, sizeof(buf));
-				if(file_exist(buf) || node_find(node->graph, buf))
+				if(node_findfile(node->graph, buf, &depnode, &timestamp))
 				{
 					found = 1;
 					break;
@@ -291,9 +317,13 @@ static int dependency_cpp_callback(struct NODE *node, void *user, const char *fi
 	if(found)
 	{
 		path_normalize(buf);
-		depnode = node_add_dependency(node, buf);
 		if(!depnode)
+			node_create(&depnode, node->graph, buf, NULL, timestamp);
+		if(node_add_dependency (node, depnode) == NULL)
 			return 2;
+
+		if(!depnode)
+			return 3;
 	
 		/* do the dependency walk */
 		if(!depnode->depchecked)
@@ -301,7 +331,7 @@ static int dependency_cpp_callback(struct NODE *node, void *user, const char *fi
 			recurseinfo.paths = depinfo->paths;
 			recurseinfo.context = depinfo->context;
 			if(dependency_cpp_run(depinfo->context, depnode, dependency_cpp_callback, &recurseinfo) != 0)
-				return 3;
+				return 4;
 		}
 	}
 		
@@ -341,19 +371,24 @@ int lf_add_dependency_cpp(lua_State *L)
 {
 	struct CONTEXT *context;
 	struct DEFERRED *deferred;
+	struct NODE * node;
 	int n = lua_gettop(L);
 	
 	if(n != 1)
-		luaL_error(L, "add_dependency_cpp_set: incorrect number of arguments");
+		luaL_error(L, "add_dependency_cpp: incorrect number of arguments");
 	luaL_checkstring(L,1);
 	
 	context = context_get_pointer(L);
-	
+	node = node_find(context->graph, lua_tostring(L,1));
+
+	if(!node)
+		luaL_error(L, "add_dependency_cpp: couldn't find node with name '%s'", lua_tostring(L,1));
+
 	deferred = (struct DEFERRED *)mem_allocate(context->deferredheap, sizeof(struct DEFERRED));
-	deferred->node = node_find(context->graph, lua_tostring(L,1));
+	deferred->node = node;
 	deferred->user = current_includepaths;
 	deferred->run = dependency_cpp_do_run;
-	deferred->next = context->firstdeferred;
-	context->firstdeferred = deferred;
+	deferred->next = context->firstdeferred_cpp;
+	context->firstdeferred_cpp = deferred;
 	return 0;
 }
