@@ -1,5 +1,8 @@
 /* platforms */
 
+#ifndef FILE_PLATFORM_H
+#define FILE_PLATFORM_H
+
 /* windows family */
 #if defined(WIN64) || defined(_WIN64)
 	/* hmm, is this IA64 or x86-64? */
@@ -20,7 +23,7 @@
 #endif
 
 /* unix family */
-#if defined(__FreeBSD__)
+#if defined(__FreeBSD__) || defined(__FreeBSD_kernel__)
 	#define BAM_FAMILY_UNIX
 	#define BAM_FAMILY_STRING "unix"
 	#define BAM_PLATFORM_FREEBSD
@@ -34,6 +37,13 @@
 	#define BAM_PLATFORM_STRING "openbsd"
 #endif
 
+#if defined(__NetBSD__)
+	#define BAM_FAMILY_UNIX
+	#define BAM_FAMILY_STRING "unix"
+	#define BAM_PLATFORM_NETBSD
+	#define BAM_PLATFORM_STRING "netbsd"
+#endif
+
 #if defined(__LINUX__) || defined(__linux__)
 	#define BAM_FAMILY_UNIX
 	#define BAM_FAMILY_STRING "unix"
@@ -55,13 +65,27 @@
 	#define BAM_PLATFORM_STRING "macosx"
 #endif
 
+#if defined(_AIX)
+	#define BAM_FAMILY_UNIX
+	#define BAM_FAMILY_STRING "unix"
+	#define BAM_PLATFORM_AIX
+	#define BAM_PLATFORM_STRING "aix"
+#endif
+
 #if defined(__sun)
 	#define BAM_FAMILY_UNIX
 	#define BAM_FAMILY_STRING "unix"
-	#define BAM_PLATFROM_SOLARIS
+	#define BAM_PLATFORM_SOLARIS
 	#define BAM_PLATFORM_STRING "solaris"
 #endif
 
+#if defined(__hpux)
+	#define BAM_FAMILY_UNIX
+	#define BAM_FAMILY_STRING "unix"
+	#define BAM_PLATFORM_HPUX
+	#define BAM_PLATFORM_STRING "hpux"
+#endif
+
 /* beos family */
 #if defined(__BeOS) || defined(__BEOS__)
 	#define BAM_FAMILY_BEOS
@@ -99,11 +123,16 @@
 	#define BAM_ARCH_STRING "ppc"
 #endif
 
-#if defined(__sparc__)
+#if defined(__sparc)
 	#define BAM_ARCH_SPARC
 	#define BAM_ARCH_STRING "sparc"
 #endif
 
+#if defined(__arm__) || defined(__ARM_ARCH_7A__)
+	#define BAM_ARCH_ARM7
+	#define BAM_ARCH_STRING "armv7l"
+#endif
+
 
 #ifndef BAM_FAMILY_STRING
 #define BAM_FAMILY_STRING "unknown"
@@ -116,3 +145,5 @@
 #ifndef BAM_ARCH_STRING
 #define BAM_ARCH_STRING "unknown"
 #endif
+
+#endif /* FILE_PLATFORM_H */
