@@ -1,3 +1,5 @@
+#include "support.h"
+
 /* session - holds all the settings for this bam session */
 struct SESSION
 {
@@ -7,15 +9,23 @@ struct SESSION
 	int verbose;
 	int simpleoutput;
 	
-	unsigned int cache_hash;
+	hash_t cache_hash;
 
+	/* lua options */
 	int lua_backtrace;
 	int lua_locals;
 	
+	/* report options */
 	int report_color;
 	int report_bar;
 	int report_steps;
-	
+
+	FILE *eventlog;
+	int eventlogflush;
+
+	/* windows options */
+	int win_msvcmode;
+
 	volatile int abort; /* raised to 1 when it's time to give up */
 };
 
