@@ -1,6 +1,4 @@
 #include <string.h>
-#include <lua.h>
-#include <lauxlib.h>
 #include <stdlib.h>
 #include <stdio.h>
 #include <time.h>
@@ -26,8 +24,10 @@
 	#include <sys/utime.h>
 	#include <signal.h>
 	#include <direct.h> /* _mkdir */
+
+	#include <Aclapi.h> /* for protect_process */
 	
-	static void list_directory(const char *path, void (*callback)(const char *filename, int dir, void *user), void *user)
+	void file_listdirectory(const char *path, void (*callback)(const char *filename, int dir, void *user), void *user)
 	{
 		WIN32_FIND_DATA finddata;
 		HANDLE handle;
@@ -64,27 +64,162 @@
 		FindClose(handle);
 	}
 
+	static HANDLE singleton_mutex = 0;
+	static DWORD observe_pid = 0;
+	static void observer_thread(void *u)
+	{
+		HANDLE process = OpenProcess(PROCESS_QUERY_INFORMATION, FALSE, observe_pid);
+		DWORD errcode = -1;
+
+		while(1) {
+			WaitForSingleObject(process, INFINITE);
+			if(GetExitCodeProcess(process, &errcode) != 0) {
+				if(errcode != STILL_ACTIVE) {
+					printf("observed process %d died, aborting!\n", observe_pid);
+					session.abort = 1;
+					break;
+				}
+			}
+			Sleep(100);
+		}
+
+		CloseHandle(process);
+	}
+
 	/* signals. should be moved to platform.c or similar? */
 	void install_signals(void (*abortsignal)(int))
 	{
+		/* abortsignal_func = abortsignal ;*/
 		signal(SIGINT, abortsignal);
 		signal(SIGBREAK, abortsignal);
 	}
 
+	static const int protect_process()
+	{
+		HANDLE hProcess = GetCurrentProcess();
+		EXPLICIT_ACCESS denyAccess = {0};
+		DWORD dwAccessPermissions = GENERIC_WRITE|PROCESS_ALL_ACCESS|WRITE_DAC|DELETE|WRITE_OWNER|READ_CONTROL;
+		PACL pTempDacl = NULL;
+		DWORD dwErr = 0;
+		BuildExplicitAccessWithName( &denyAccess, "CURRENT_USER", dwAccessPermissions, DENY_ACCESS, NO_INHERITANCE );
+		dwErr = SetEntriesInAcl( 1, &denyAccess, NULL, &pTempDacl );
+		/* check dwErr... */
+		dwErr = SetSecurityInfo( hProcess, SE_KERNEL_OBJECT, DACL_SECURITY_INFORMATION, NULL, NULL, pTempDacl, NULL );
+		/* check dwErr... */
+		LocalFree( pTempDacl );
+		CloseHandle( hProcess );
+		return dwErr == ERROR_SUCCESS;
+	}
+
 	static CRITICAL_SECTION criticalsection;
 	void platform_init()
 	{
-		InitializeCriticalSection(&criticalsection);
-		
+		char buffer[512] = {0};
+
 		/* this environment variable is set by Microsoft Visual Studio
 			when building. It causes cl.exe to redirect it's output to
 			the specified pipe id. this causes loads of problems with
 			output.
 		*/ 
 		SetEnvironmentVariable("VS_UNICODE_OUTPUT", NULL);
+
+		/* check if we are being spawned by bam in msvc mode so we should be singleton */
+		if(GetEnvironmentVariable("BAM_SINGLETON", buffer, sizeof(buffer)-1))
+		{
+			DWORD ret;
+			SetEnvironmentVariable("BAM_SINGLETON", NULL);
+
+			singleton_mutex = CreateMutex(NULL, FALSE, "Global\\bam_singleton_mutex");
+			if(!singleton_mutex)
+			{
+				printf("bam is already running, wait for it to finish and then try again 1\n");
+				exit(1);
+			}
+
+			while(1)
+			{
+				ret = WaitForSingleObject( singleton_mutex, 100 );
+				if(ret == WAIT_OBJECT_0)
+					break;
+				else
+				{
+					/* printf("bam is already running, wait for it to finish and then try again 2 (%d)\n", ret); */
+					printf("bam is already running, waiting for it to finish\n"); fflush(stdout);
+					Sleep(1000);
+				}
+			}
+		}
+
+		/* this environment variable can be setup so that bam can observe if
+			a specific process dies and abort the building if it does. This
+			is used in conjunction with Microsoft Visual Studio to make sure
+			that it doesn't kill processes wildly.
+		*/
+		if(GetEnvironmentVariable("BAM_OBSERVE_PID", buffer, sizeof(buffer)-1))
+		{
+			observe_pid = atoi(buffer);
+			threads_create(observer_thread, NULL);
+
+			/* protect process from being killed */
+			protect_process();
+		}
+		else
+		{
+			if( session.win_msvcmode )
+			{
+				DWORD errcode;
+				PROCESS_INFORMATION pi;
+				STARTUPINFO si;
+
+				/* setup ourself to be watched */
+				sprintf(buffer, "%d", GetCurrentProcessId());
+				SetEnvironmentVariable("BAM_OBSERVE_PID", buffer);
+
+				/* signal that we want to be singleton */
+				SetEnvironmentVariable("BAM_SINGLETON", "1");
+
+				/* init structs and create process */
+				ZeroMemory( &si, sizeof(si) );
+				si.cb = sizeof(si);
+				ZeroMemory( &pi, sizeof(pi) );
+
+				if( CreateProcess(
+					NULL,      /* No module name (use command line) */
+					GetCommandLine(),   /* Command line */
+					NULL,      /* Process handle not inheritable */
+					NULL,      /* Thread handle not inheritable */
+					TRUE,      /* Set handle inheritance to FALSE */
+					0,         /* No creation flags */
+					NULL,      /* Use parent's environment block */
+					NULL,      /* Use parent's starting directory */
+					&si,       /* Pointer to STARTUPINFO structure */
+					&pi        /* Pointer to PROCESS_INFORMATION structure */
+				) ) {
+					/* wait for the child to complete */
+					WaitForSingleObject(pi.hProcess, INFINITE);
+					GetExitCodeProcess(pi.hProcess, &errcode);
+					CloseHandle(pi.hProcess);
+					CloseHandle(pi.hThread);
+					exit(errcode);
+				} else {
+					printf("failed to spawn new bam process in msvc mode\n");
+					printf("%s\n", GetCommandLine());
+					exit(1);
+				}
+			}
+		}
+
+		InitializeCriticalSection(&criticalsection);
 	}
 	
-	void platform_shutdown() { DeleteCriticalSection(&criticalsection); }
+	void platform_shutdown()
+	{
+		if(singleton_mutex)
+			ReleaseMutex(singleton_mutex);
+		CloseHandle(singleton_mutex);
+		DeleteCriticalSection(&criticalsection);
+	}
+
 	void criticalsection_enter() { EnterCriticalSection(&criticalsection); }
 	void criticalsection_leave() { LeaveCriticalSection(&criticalsection); }
 
@@ -102,31 +237,39 @@
 	{
 		Sleep(1);
 	}
-	
-	PLUGINFUNC plugin_load(const char *filename)
-	{
-		char buffer[MAX_PATH_LENGTH];
-		HMODULE handle;
-		FARPROC func;
-		
-		_snprintf(buffer, sizeof(buffer), "%s.dll", filename);
 
-		handle = LoadLibrary(buffer);
-		if(handle == NULL)
-		{
-			fprintf(stderr, "error loading plugin '%s'\n", buffer);
-			return NULL;
-		}
-		
-		func = GetProcAddress(handle, "plugin_main");
-		if(func == NULL)
-		{
-			CloseHandle(handle);
-			fprintf(stderr, "error fetching plugin main from '%s'\n", buffer);
-			return NULL;
-		}
-		
-		return (PLUGINFUNC)func;
+
+	void threads_sleep(int milliseconds)
+	{
+		Sleep(milliseconds);
+	}
+	
+	int threads_corecount()
+	{
+		SYSTEM_INFO sysinfo;
+		GetSystemInfo(&sysinfo);
+		if(sysinfo.dwNumberOfProcessors >= 1)
+			return sysinfo.dwNumberOfProcessors;
+		return 1;
+	}
+
+
+	int64 time_get()
+	{
+		static int64 last = 0;
+		int64 t;
+		QueryPerformanceCounter((PLARGE_INTEGER)&t);
+		if(t<last) /* for some reason, QPC can return values in the past */
+			return last;
+		last = t;
+		return t;
+	}
+
+	int64 time_freq()
+	{
+		int64 t;
+		QueryPerformanceFrequency((PLARGE_INTEGER)&t);
+		return t;
 	}
 
 #else
@@ -142,15 +285,24 @@
 	#include <sys/types.h>
 	#include <sys/signal.h>
 	#include <sys/stat.h>
+	#include <sys/wait.h> 
 	#include <utime.h>
 	#include <pthread.h>
+	#include <signal.h>
+
+	#include <sys/time.h>
+
+#ifdef BAM_PLATFORM_MACOSX
+	#include <sys/param.h>
+	#include <sys/sysctl.h>
+#endif
 
 /* disable D_TYPE_HACK if we don't have support for it */
 #if !defined(DT_DIR) || !defined(DT_UNKNOWN)
 	#undef D_TYPE_HACK
 #endif
 
-	static void list_directory(const char *path, void (*callback)(const char *filename, int dir, void *user), void *user)
+	void file_listdirectory(const char *path, void (*callback)(const char *filename, int dir, void *user), void *user)
 	{
 		DIR *dir;
 		struct dirent *entry;
@@ -240,46 +392,57 @@
 	{
 		sched_yield();
 	}
-	
-	PLUGINFUNC plugin_load(const char *filename)
+
+	void threads_sleep(int milliseconds)
 	{
-		char buffer[MAX_PATH_LENGTH];
-		const char *error;
-		void *handle;
-		union
-		{
-			PLUGINFUNC func;
-			void *ptr;
-		} func;
-		
-		if(strlen(filename) > sizeof(buffer) - 10)
-			return (PLUGINFUNC)0;
-		
-		strcpy(buffer, "./");
-		strcat(buffer, filename);
-		strcpy(buffer, ".so");
+		usleep(milliseconds*1000);
+	}
 
-		handle = dlopen(buffer, RTLD_LAZY);
-		if(!handle)
-		{
-			fputs(dlerror(), stderr);
-			fputs("\n", stderr);
-			
-			return NULL;
-		}
-		
-		func.ptr = dlsym(handle, "plugin_main");
-		error = dlerror();
+	int threads_corecount()
+	{
+#ifdef BAM_PLATFORM_MACOSX
+		int nm[2] = {CTL_HW, HW_AVAILCPU};
+		size_t len = 4;
+		uint32_t count;
 
-		if(error)
+		sysctl(nm, 2, &count, &len, NULL, 0);
+
+		if(count < 1)
 		{
-			fputs(error, stderr);
-			fputs("\n", stderr);
-			dlclose(handle);
-			return NULL;
+			nm[1] = HW_NCPU;
+			sysctl(nm, 2, &count, &len, NULL, 0);
+			if(count < 1) { count = 1; }
 		}
-		
-		return func.func;
+		if(count >= 1)
+			return count;
+		return 1;
+#elif defined(BAM_PLATFORM_HPUX)
+#include <sys/pstat.h>
+	struct pst_dynamic psd;
+
+	if (!pstat_getdynamic(&psd, sizeof(psd), (size_t)1, 0))
+	{
+		return psd.psd_proc_cnt;
+	}
+	return 1;
+#else
+	    int count = sysconf(_SC_NPROCESSORS_ONLN);
+	    if(count >= 1)
+	    	return count;
+	    return 1;
+#endif
+	}
+
+	int64 time_get()
+	{
+		struct timeval val;
+		gettimeofday(&val, NULL);
+		return (int64)val.tv_sec*(int64)1000000+(int64)val.tv_usec;
+	}
+
+	int64 time_freq()
+	{
+		return 1000000;
 	}
 		
 #endif
@@ -304,6 +467,23 @@ time_t file_timestamp(const char *filename)
 #endif
 }
 
+int file_isregular(const char *filename)
+{
+#ifdef BAM_FAMILY_WINDOWS
+	struct stat s;
+	if(stat(filename, &s) == 0)
+	{
+		if (s.st_mode&_S_IFREG)
+			return 1;
+	}
+#else
+	struct stat s;
+	if(stat(filename, &s) == 0)
+		return S_ISREG(s.st_mode);
+#endif
+	return 0;
+}
+
 int file_createdir(const char *path)
 {
 	int r;
@@ -320,7 +500,30 @@ int file_createdir(const char *path)
 void file_touch(const char *filename)
 {
 #ifdef BAM_FAMILY_WINDOWS
-	_utime(filename, NULL);
+	/*
+		_utime under windows seem to contain a bug that doesn't release the file handle in a timly fashion.
+		This implementation is basiclly the same but smaller and less cruft.
+	*/
+	HANDLE handle;
+	FILETIME ft;
+	SYSTEMTIME st;
+
+	handle = CreateFile(
+		filename,
+		GENERIC_READ | GENERIC_WRITE,
+		FILE_SHARE_WRITE | FILE_SHARE_READ | FILE_SHARE_DELETE,
+		NULL,
+		OPEN_EXISTING,
+		FILE_ATTRIBUTE_NORMAL,
+		NULL);
+
+	if(handle != INVALID_HANDLE_VALUE)
+	{
+		GetSystemTime(&st);
+		SystemTimeToFileTime(&st, &ft);
+		SetFileTime(handle, (LPFILETIME)NULL, (LPFILETIME)NULL, &ft);
+		CloseHandle(handle);
+	}
 #else
 	utime(filename, NULL);
 #endif
@@ -342,11 +545,41 @@ static void passthru(FILE *fp)
 }
 #endif
 
+#if defined(BAM_FAMILY_WINDOWS) || defined(BAM_PLATFORM_CYGWIN)
+/* forward declaration */
+FILE *_popen(const char *, const char *);
+int _pclose(FILE *);
+#endif
+
 int run_command(const char *cmd, const char *filter)
 {
 	int ret;
+	
 #ifdef BAM_FAMILY_WINDOWS
-	FILE *fp = _popen(cmd, "r");
+	/* windows has a buggy command line parser. I takes the first and
+	last '"' and removes them which causes problems in cases like this:
+	
+		"C:\t t\test.bat" 1 2 "3" 4 5
+		
+	Which will yeild:
+
+		C:\t t\test.bat" 1 2 "3 4 5
+		
+	The work around is to encase the command line with '"' that it can remove.
+	
+	*/
+	char buf[32*1024];
+	FILE *fp;
+	size_t len = strlen(cmd);
+	if(len > sizeof(buf)-3) /* 2 for '"' and 1 for '\0' */
+		return -1;
+	buf[0] = '"';
+	memcpy(buf+1, cmd, len);
+	buf[len+1] = '"';
+	buf[len+2] = 0;
+	
+	/* open the command line */
+	fp = _popen(buf, "r");
 
 	if(!fp)
 		return -1;
@@ -424,205 +657,68 @@ int run_command(const char *cmd, const char *filter)
 	ret = _pclose(fp);
 #else
 	ret = system(cmd);
+	if(WIFSIGNALED(ret))
+		raise(SIGINT);
 #endif
 	if(session.verbose)
 		printf("%s: ret=%d %s\n", session.name, ret, cmd);
 	return ret;
 }
 
-/* general */
-int file_exist(const char *filename)
+/* like file_createdir, but automatically creates all top-level directories needed
+	If you feed it "output/somefiles/output.o" it will create "output/somefiles"
+*/	
+int file_createpath(const char *output_name)
 {
-	struct stat s;
-	if(stat(filename, &s) == 0)
-		return 1;
+	char buffer[MAX_PATH_LENGTH];
+	int i;
+	char t;
+	
+	/* fish out the directory */
+	if(path_directory(output_name, buffer, sizeof(buffer)) != 0)
+	{
+		fprintf(stderr, "path error: %s\n", buffer);
+		return -1;
+	}
+	
+	/* no directory in path */
+	if(buffer[0] == 0)
+		return 0;
+	
+	/* check if we need to do a deep walk */
+	if(file_createdir(buffer) == 0)
+		return 0;
+	
+	/* create dir by doing a deep walk */
+	i = 0;
+	while(1)
+	{
+		if((buffer[i] == '/') || (buffer[i] == 0))
+		{
+			/* insert null terminator */
+			t = buffer[i];
+			buffer[i] = 0;
+			
+			if(file_createdir(buffer) != 0)
+			{
+				fprintf(stderr, "path error2: %s\n", buffer);
+				return -1;
+			}
+			
+			/* restore the path */
+			buffer[i] = t;
+		}
+		
+		if(buffer[i] == 0)
+			break;
+		
+		i++;
+	}
+	
+	/* return success */
 	return 0;
 }
 
-/* list directory functionallity */
-typedef struct
-{
-	lua_State *lua;
-	int i;
-} LISTDIR_CALLBACK_INFO;
-
-static void listdir_callback(const char *filename, int dir, void *user)
-{
-	LISTDIR_CALLBACK_INFO *info = (LISTDIR_CALLBACK_INFO *)user;
-	lua_pushstring(info->lua, filename);
-	lua_rawseti(info->lua, -2, info->i++);
-}
-
-int lf_listdir(lua_State *L)
-{
-	LISTDIR_CALLBACK_INFO info;
-	info.lua = L;
-	info.i = 1;
-	
-	/* create the table */
-	lua_newtable(L);
-
-	/* add all the entries */
-	if(strlen(lua_tostring(L, 1)) < 1)
-		list_directory(context_get_path(L), listdir_callback, &info);
-	else
-	{
-		char buffer[1024];
-		path_join(context_get_path(L), -1, lua_tostring(L,1), -1, buffer, sizeof(buffer));
-		list_directory(buffer, listdir_callback, &info);
-	}
-
-	return 1;
-}
-
-/* collect functionallity */
-enum
-{
-	COLLECTFLAG_FILES=1,
-	COLLECTFLAG_DIRS=2,
-	COLLECTFLAG_HIDDEN=4,
-	COLLECTFLAG_RECURSIVE=8
-};
-
-typedef struct
-{
-	int path_len;
-	const char *start_str;
-	int start_len;
-	
-	const char *end_str;
-	int end_len;
-	
-	lua_State *lua;
-	int i;
-	int flags;
-} COLLECT_CALLBACK_INFO;
-
-static void run_collect(COLLECT_CALLBACK_INFO *info, const char *input);
-
-static void collect_callback(const char *filename, int dir, void *user)
-{
-	COLLECT_CALLBACK_INFO *info = (COLLECT_CALLBACK_INFO *)user;
-	const char *no_pathed = filename + info->path_len;
-	int no_pathed_len = strlen(no_pathed);
-
-	/* don't process . and .. paths */
-	if(filename[0] == '.')
-	{
-		if(filename[1] == 0)
-			return;
-		if(filename[1] == '.' && filename[2] == 0)
-			return;
-	}
-	
-	/* don't process hidden stuff if not wanted */
-	if(no_pathed[0] == '.' && !(info->flags&COLLECTFLAG_HIDDEN))
-		return;
-
-	do
-	{
-		/* check end */
-		if(info->end_len > no_pathed_len || strcmp(no_pathed+no_pathed_len-info->end_len, info->end_str))
-			break;
-
-		/* check start */
-		if(info->start_len && strncmp(no_pathed, info->start_str, info->start_len))
-			break;
-		
-		/* check dir vs search param */
-		if(!dir && info->flags&COLLECTFLAG_DIRS)
-			break;
-		
-		if(dir && info->flags&COLLECTFLAG_FILES)
-			break;
-			
-		/* all criterias met, push the result */
-		lua_pushstring(info->lua, filename);
-		lua_rawseti(info->lua, -2, info->i++);
-	} while(0);
-	
-	/* recurse */
-	if(dir && info->flags&COLLECTFLAG_RECURSIVE)
-	{
-		char recursepath[1024];
-		COLLECT_CALLBACK_INFO recurseinfo = *info;
-		strcpy(recursepath, filename);
-		strcat(recursepath, "/");
-		strcat(recursepath, info->start_str);
-		run_collect(&recurseinfo, recursepath);
-		info->i = recurseinfo.i;
-	}	
-}
-
-static void run_collect(COLLECT_CALLBACK_INFO *info, const char *input)
-{
-	char dir[1024];
-	int dirlen = 0;
-	
-	/* get the directory */
-	path_directory(input, dir, sizeof(dir));
-	dirlen = strlen(dir);
-	info->path_len = dirlen+1;
-	
-	/* set the start string */
-	if(dirlen)
-		info->start_str = input + dirlen + 1;
-	else
-		info->start_str = input;
-		
-	for(info->start_len = 0; info->start_str[info->start_len]; info->start_len++)
-	{
-		if(info->start_str[info->start_len] == '*')
-			break;
-	}
-	
-	/* set the end string */
-	if(info->start_str[info->start_len])
-		info->end_str = info->start_str + info->start_len + 1;
-	else
-		info->end_str = info->start_str + info->start_len;
-	info->end_len = strlen(info->end_str);
-	
-	/* search the path */
-	list_directory(dir, collect_callback, info);	
-}
-
-static int collect(lua_State *L, int flags)
-{
-	int n = lua_gettop(L);
-	int i;
-	COLLECT_CALLBACK_INFO info;
-	
-	if(n < 1)
-		luaL_error(L, "collect: incorrect number of arguments");
-
-	/* create the table */
-	lua_newtable(L);
-
-	/* set common info */		
-	info.lua = L;
-	info.i = 1;
-	info.flags = flags;
-
-	/* start processing the input strings */
-	for(i = 1; i <= n; i++)
-	{
-		const char *input = lua_tostring(L, i);
-		
-		if(!input)
-			continue;
-			
-		run_collect(&info, input);
-	}
-	
-	return 1;
-}
-
-int lf_collect(lua_State *L) { return collect(L, COLLECTFLAG_FILES); }
-int lf_collectrecursive(lua_State *L) { return collect(L, COLLECTFLAG_FILES|COLLECTFLAG_RECURSIVE); }
-int lf_collectdirs(lua_State *L) { return collect(L, COLLECTFLAG_DIRS); }
-int lf_collectdirsrecursive(lua_State *L) { return collect(L, COLLECTFLAG_DIRS|COLLECTFLAG_RECURSIVE); }
-
 /* */
 #ifdef BAM_FAMILY_WINDOWS
 /* on windows, we need to handle that filenames with mixed casing are the same.
@@ -648,24 +744,61 @@ static const unsigned char tolower_table[256] = {
 224, 225, 226, 227, 228, 229, 230, 231, 232, 233, 234, 235, 236, 237, 238, 239,
 240, 241, 242, 243, 244, 245, 246, 247, 248, 249, 250, 251, 252, 253, 254, 255};
 
-unsigned int string_hash_add(unsigned int h, const char *str_in)
+hash_t string_hash_add(hash_t h, const char *str_in)
 {
 	const unsigned char *str = (const unsigned char *)str_in;
 	for (; *str; str++)
-		h = 31*h + tolower_table[*str];
+		h = (33*h) ^ tolower_table[*str];
 	return h;
 }
 #else
 /* normal unix version */
-unsigned int string_hash_add(unsigned int h, const char *str)
+hash_t string_hash_add(hash_t h, const char *str)
 {
 	for (; *str; str++)
-		h = 31*h + *str;
+		h = (33*h) ^ *str;
 	return h;
 }
 #endif
 
-unsigned int string_hash(const char *str_in)
+hash_t string_hash(const char *str_in)
 {
-	return string_hash_add(0, str_in);
+	return string_hash_add(5381, str_in);
+}
+
+void string_hash_tostr(hash_t value, char *output)
+{
+	sprintf(output, "%08x%08x", (unsigned)(value>>32), (unsigned)(value&0xffffffff));
+}
+
+
+static int64 starttime = 0;
+
+static void event_log(int thread, const char *type, const char *name, const char *data)
+{
+	double t;
+	if(session.eventlog == NULL)
+		return;
+
+	if(starttime == 0)
+		starttime = time_get();
+
+	if(data == NULL)
+		data = "";
+
+	t = (time_get() - starttime) / (double)time_freq();
+	fprintf(session.eventlog, "%d %f %s %s: %s\n", thread, t, type, name, data);
+
+	if(session.eventlogflush)
+		fflush(session.eventlog);
+}
+
+void event_begin(int thread, const char *name, const char *data)
+{
+	event_log(thread, "begin", name, data);
+}
+
+void event_end(int thread, const char *name, const char *data)
+{
+	event_log(thread, "end", name, data);
 }
